from typing import Dict, Any, Union

from sybil.typing import Evaluator


class Lexeme(str):
    """
    Where needed, this can store both the text of the lexeme
    and it's line offset relative to the line number of the example
    that contains it.
    """

    def __new__(cls, text: str, offset, line_offset: int):
        return str.__new__(cls, text)

    def __init__(self, text: str, offset, line_offset: int):
        self.text, self.offset, self.line_offset = text, offset, line_offset


class LexedRegion:
    """
    A region that has been lexed from a source language but not yet
    parsed or assigned an :any:`Evaluator`.
    """

    def __init__(self, start: int, end: int, lexemes: Dict[str, Union[str, Lexeme]]):
        #: The start of this lexed region within the document's :attr:`~sybil.Document.text`.
        self.start: int = start
        #: The end of this lexed region within the document's :attr:`~sybil.Document.text`.
        self.end: int = end
        #: The lexemes extracted from the region.
        self.lexemes: Dict[str, Union[str, Lexeme]] = lexemes


class Region:
    """
    Parsers should yield instances of this class for each example they
    discover in a documentation source file.
    
    :param start: 
        The character position at which the example starts in the
        :class:`~sybil.document.Document`.
    
    :param end: 
        The character position at which the example ends in the
        :class:`~sybil.document.Document`.
    
    :param parsed: 
        The parsed version of the example.
    
    :param evaluator: 
        The callable to use to evaluate this example and check if it is
        as it should be.
    """

    def __init__(self, start: int, end: int, parsed: Any, evaluator: Evaluator):
        #: The start of this region within the document's :attr:`~sybil.Document.text`.
        self.start: int = start
        #: The end of this region within the document's :attr:`~sybil.Document.text`.
        self.end: int = end
        #: The parsed version of this region. This only needs to have meaning to
        #: the :attr:`evaluator`.
        self.parsed: Any = parsed
        #: The :any:`Evaluator` for this region.
        self.evaluator: Evaluator = evaluator

    def __repr__(self) -> str:
        return '<Region start={} end={} {!r}>'.format(
            self.start, self.end, self.evaluator
        )

    def __lt__(self, other):
        assert isinstance(other, type(self)), f"{type(other)} not supported for <"
        assert self.start == other.start  # This is where this may happen, if not something weird
        return True

    def adjust(self, lexed: Union['Region', 'LexedRegion'], lexeme: Lexeme):
        """
        Adjust the start and end of this region based on the provided :class:`Lexeme`
        and :class:`LexedRegion` or :class:`Region` that lexeme came from.
        """
        self.start += (lexed.start + lexeme.offset)
        self.end += lexed.start



import re
from typing import Dict

from sybil.parsers.abstract.lexers import BlockLexer

START_PATTERN_TEMPLATE =(
    r'^(?P<prefix>[ \t]*)\.\.\s*(?P<directive>{directive})'
    r'{delimiter}\s*'
    r'(?P<arguments>[\w-]+\b)?'
    r'(?:\s*\:[\w-]+\:.*\n)*'
    r'(?:\s*\n)*\n'
)

END_PATTERN_TEMPLATE = '(\n\\Z|\n[ \t]{{0,{len_prefix}}}(?=\\S))'


class DirectiveLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for ReST directives that extracts the
    following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.
    """

    delimiter = '::'

    def __init__(self, directive: str, arguments: str = '', mapping: Dict[str, str] = None):
        """
        A lexer for ReST directives.
        Both ``directive`` and ``arguments`` are regex patterns.
        """
        super().__init__(
            start_pattern=re.compile(
                START_PATTERN_TEMPLATE.format(
                    directive=directive,
                    delimiter=self.delimiter,
                    arguments=arguments
                ),
                re.MULTILINE
            ),
            end_pattern_template=END_PATTERN_TEMPLATE,
            mapping=mapping,
        )


class DirectiveInCommentLexer(DirectiveLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for faux ReST directives in comments
    such as:

    .. code-block:: rest

        .. not-really-a-directive: some-argument

          Source here...

    It extracts the following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.
    """

    # This is the pattern used for invisible code blocks and the like.
    delimiter = ':'



import re
from typing import Dict

from sybil.parsers.abstract.lexers import BlockLexer

CODEBLOCK_START_TEMPLATE = r"^(?P<prefix>[ \t]*)```(?P<language>{language})$\n"
CODEBLOCK_END_TEMPLATE = r"(?<=\n){prefix}```\n"


class FencedCodeBlockLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for Markdown fenced code blocks.

    The following lexemes are extracted:

    - ``language`` as a  :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.


    :param language:
        a :class:`str` containing a regular expression pattern to match language names.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their
        values. Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.

    """

    def __init__(self, language: str, mapping: Dict[str, str] = None):
        super().__init__(
            start_pattern=re.compile(CODEBLOCK_START_TEMPLATE.format(language=language)),
            end_pattern_template=CODEBLOCK_END_TEMPLATE,
            mapping=mapping,
        )
        self.start_pattern = re.compile(
            CODEBLOCK_START_TEMPLATE.format(language=language),
            re.MULTILINE
        )


DIRECTIVE_START_TEMPLATE = (
    r"^(?P<prefix>[ \t]*)```\{{(?P<directive>{directive})}} ?(?P<arguments>{arguments})$\n"
    r"(\1---\n(.+\n)*\1---\n)?"
)


class DirectiveLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for MyST directives such as:

    .. code-block:: markdown

        ```{directivename} arguments
        ---
        key1: val1
        key2: val2
        ---
        This is
        directive content
        ```

    The following lexemes are extracted:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their
        values. Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.

    """

    def __init__(self, directive: str, arguments: str = '', mapping: Dict[str, str] = None):
        super().__init__(
            start_pattern=re.compile(
                DIRECTIVE_START_TEMPLATE.format(directive=directive, arguments=arguments),
                re.MULTILINE
            ),
            end_pattern_template=CODEBLOCK_END_TEMPLATE,
            mapping=mapping,
        )


DIRECTIVE_IN_PERCENT_COMMENT_START = (
    r"^(?P<prefix>[ \t]*%)[ \t]*(?P<directive>{directive}):[ \t]*(?P<arguments>{arguments})$\n"
)
DIRECTIVE_IN_PERCENT_COMMENT_END = '(?<=\n)(?!{prefix})'


class DirectiveInPercentCommentLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for faux MyST directives in
    %-style Markdown comments such as:

    .. code-block:: markdown

        ; not-really-a-directive: some-argument
        ;
        ;     Source here...

    It extracts the following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.
    """

    def __init__(self, directive: str, arguments: str = '', mapping: Dict[str, str] = None):
        super().__init__(
            start_pattern=re.compile(
                DIRECTIVE_IN_PERCENT_COMMENT_START.format(directive=directive, arguments=arguments),
                re.MULTILINE
            ),
            end_pattern_template=DIRECTIVE_IN_PERCENT_COMMENT_END,
            mapping=mapping,
        )


DIRECTIVE_IN_HTML_COMMENT_START = (
    r"^(?P<prefix>[ \t]*)<!---[ \t]*(?P<directive>{directive}):[ \t]*(?P<arguments>{arguments})$\n"
)
DIRECTIVE_IN_HTML_COMMENT_END = '(?<=\n){prefix}--->'


class DirectiveInHTMLCommentLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for faux MyST directives in
    HTML-style Markdown comments such as:

    .. code-block:: markdown

        <!--- not-really-a-directive: some-argument

            Source here...

        --->

    It extracts the following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.
    """

    def __init__(self, directive: str, arguments: str = '', mapping: Dict[str, str] = None):
        super().__init__(
            start_pattern=re.compile(
                DIRECTIVE_IN_HTML_COMMENT_START.format(directive=directive, arguments=arguments),
                re.MULTILINE
            ),
            end_pattern_template=DIRECTIVE_IN_HTML_COMMENT_END,
            mapping=mapping,
        )



from typing import Callable, TYPE_CHECKING, Iterable, Optional, Any

if TYPE_CHECKING:
    import sybil


#: The signature for an evaluator. See :ref:`developing-parsers`.
Evaluator = Callable[['sybil.Example'], Optional[str]]

#: The signature for a lexer. See :ref:`developing-parsers`.
Lexer = Callable[['sybil.Document'], Iterable['sybil.LexedRegion']]

#: The signature for a parser. See :ref:`developing-parsers`.
Parser = Callable[['sybil.Document'], Iterable['sybil.Region']]



import re
import textwrap
from typing import Dict, Iterable, Pattern

from sybil import Document
from sybil.region import LexedRegion, Lexeme


class BlockLexer:
    """
    This is a base class useful for any :any:`Lexer` that must handle block-style languages
    such as ReStructured Text or MarkDown.

    It yields a sequence of :class:`~sybil.LexedRegion` objects for each case where the
    ``start_pattern`` matches. A ``source`` :class:`~sybil.Lexeme` is created from the text between
    the end of the start pattern and the start of the end pattern.

    :param start_pattern:
        This is used to match the start of the block. Any named groups will be returned
        in the :attr:`~sybil.LexedRegion.lexemes` :class:`dict` of resulting
        :class:`~sybil.LexedRegion` objects. If a ``prefix`` named group forms
        part of the match, this will be template substituted into the ``end_pattern_template``
        before it is compiled.

    :param end_pattern_template:
        This is used to match the end of any block found by the ``start_pattern``.
        It is templated with any ``prefix`` group from the ``start_pattern`` :class:`~typing.Match`
        and ``len_prefix``, the length of that prefix, before being compiled into a
        :class:`~typing.Pattern`.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.LexedRegion` objects.
    """

    def __init__(
            self,
            start_pattern: Pattern,
            end_pattern_template: str,
            mapping: Dict[str, str] = None,
    ):
        self.start_pattern = start_pattern
        self.end_pattern_template = end_pattern_template
        self.mapping = mapping

    def __call__(self, document: Document) -> Iterable[LexedRegion]:
        for start_match in re.finditer(self.start_pattern, document.text):
            source_start = start_match.end()
            lexemes = start_match.groupdict()
            prefix = lexemes.pop('prefix', '')
            end_pattern = re.compile(self.end_pattern_template.format(
                prefix=prefix, len_prefix=len(prefix)
            ))
            end_match = end_pattern.search(document.text, source_start)
            source_end = end_match.start()
            source = document.text[source_start:source_end]
            lines = source.splitlines(keepends=True)
            stripped = ''.join(line[len(prefix):] for line in lines)
            lexemes['source'] = Lexeme(
                textwrap.dedent(stripped),
                offset=source_start-start_match.start(),
                line_offset=start_match.group(0).count('\n')-1
            )
            if self.mapping:
                lexemes = {dest: lexemes[source] for source, dest in self.mapping.items()}
            yield LexedRegion(start_match.start(), source_end, lexemes)



from itertools import chain
from typing import Iterable, Sequence, Optional

from sybil import Region, Document, Example
from sybil.typing import Evaluator, Lexer


class AbstractCodeBlockParser:
    """
    An abstract parser for use when evaluating blocks of code.

    :param lexers:
        A sequence of :any:`Lexer` objects that will be applied in turn to each
        :class:`~sybil.Document`
        that is parsed. The :class:`~sybil.LexedRegion` objects returned by these lexers must have
        both an ``arguments`` string, containing the language of the lexed region, and a
        ``source`` :class:`~sybil.Lexeme` containing the source code of the lexed region.

    :param language:
        The language that this parser should look for. Lexed regions which don't have this
        language in their ``arguments`` lexeme will be ignored.

    :param evaluator:
        The evaluator to use for evaluating code blocks in the specified language.
        You can also override the :meth:`evaluate` method below.
    """

    language: str

    def __init__(self, lexers: Sequence[Lexer], language: str = None, evaluator: Evaluator = None):
        self.lexers = lexers
        if language is not None:
            self.language = language
        assert self.language, 'language must be specified!'
        if evaluator is not None:
            self.evaluate = evaluator

    def evaluate(self, example: Example) -> Optional[str]:
        """
        The :any:`Evaluator` used for regions yields by this parser can be provided by
        implementing this method.
        """
        raise NotImplementedError

    def __call__(self, document: Document) -> Iterable[Region]:
        for lexed in chain(*(lexer(document) for lexer in self.lexers)):
            if lexed.lexemes['arguments'] == self.language:
                yield Region(lexed.start, lexed.end, lexed.lexemes['source'], self.evaluate)



from typing import Iterable

from sybil.evaluators.python import PythonEvaluator
from sybil.parsers.abstract import AbstractCodeBlockParser, DocTestStringParser
from sybil.typing import Evaluator
from .lexers import (
    DirectiveLexer, FencedCodeBlockLexer, DirectiveInPercentCommentLexer,
    DirectiveInHTMLCommentLexer
)
from ... import Document, Region
from ...evaluators.doctest import DocTestEvaluator


class CodeBlockParser(AbstractCodeBlockParser):
    """
    A :any:`Parser` for :ref:`myst-codeblock-parser` examples.

    :param language:
        The language that this parser should look for.

    :param evaluator:
        The evaluator to use for evaluating code blocks in the specified language.
        You can also override the :meth:`evaluate` method below.
    """

    def __init__(self, language: str = None, evaluator: Evaluator = None):
        super().__init__(
            [
                FencedCodeBlockLexer(
                    language=r'.+',
                    mapping={'language': 'arguments', 'source': 'source'},
                ),
                DirectiveLexer(
                    directive='code-block',
                    arguments='.+',
                ),
                DirectiveInPercentCommentLexer(
                    directive=r'(invisible-)?code(-block)?',
                    arguments='.+',
                ),
                DirectiveInHTMLCommentLexer(
                    directive=r'(invisible-)?code(-block)?',
                    arguments='.+',
                ),
            ],
            language, evaluator
        )


class PythonCodeBlockParser(CodeBlockParser):
    """
    A :any:`Parser` for Python :ref:`myst-codeblock-parser` examples.

    :param future_imports:
        An optional list of strings that will be turned into
        ``from __future__ import ...`` statements and prepended to the code
        in each of the examples found by this parser.

    :param doctest_optionflags:
        :ref:`doctest option flags<option-flags-and-directives>` to use
        when evaluating the doctest examples found by this parser.
    """

    language = 'python'

    def __init__(self, future_imports=(), doctest_optionflags=0):
        super().__init__(evaluator=PythonEvaluator(future_imports))
        self.doctest_parser = DocTestStringParser(DocTestEvaluator(doctest_optionflags))

    def __call__(self, document: Document) -> Iterable[Region]:
        for region in super().__call__(document):
            source = region.parsed
            if region.parsed.startswith('>>>'):
                for doctest_region in self.doctest_parser(source, document.path):
                    doctest_region.adjust(region, source)
                    yield doctest_region
            else:
                yield region



import re

from sybil.parsers.abstract import AbstractSkipParser


class SkipParser(AbstractSkipParser):
    """
    A :any:`Parser` for :ref:`skip <myst-skip-parser>` instructions.
    """

    pattern = re.compile(r'^[ \t]*;\s*skip:\s*(\w+)(?:\s+if(.+)$)?', re.MULTILINE)



from doctest import (
    DocTest as BaseDocTest,
    DocTestRunner as BaseDocTestRunner,
    _unittest_reportflags,
)

from sybil import Example


class DocTest(BaseDocTest):
    def __init__(self, examples, globs, name, filename, lineno, docstring):
        # do everything like regular doctests, but don't make a copy of globs
        BaseDocTest.__init__(self, examples, globs, name, filename, lineno, docstring)
        self.globs = globs


class DocTestRunner(BaseDocTestRunner):

    def __init__(self, optionflags):
        optionflags |= _unittest_reportflags
        BaseDocTestRunner.__init__(
            self,
            verbose=False,
            optionflags=optionflags,
        )

    def _failure_header(self, test, example):
        return ''


class DocTestEvaluator:
    """
    The :any:`Evaluator` to use for :class:`Regions <sybil.Region>` yielded by
    a :class:`~sybil.parsers.abstract.doctest.DocTestStringParser`.


    :param optionflags:
        :ref:`doctest option flags<option-flags-and-directives>` to use
        when evaluating examples.
    """

    def __init__(self, optionflags=0):
        self.runner = DocTestRunner(optionflags)

    def __call__(self, sybil_example: Example) -> str:
        example = sybil_example.parsed
        namespace = sybil_example.namespace
        output = []
        self.runner.run(
            DocTest([example], namespace, name=sybil_example.path,
                    filename=None, lineno=example.lineno, docstring=None),
            clear_globs=False,
            out=output.append
        )
        return ''.join(output)



import re
from typing import Iterable, Pattern

from sybil import Document, Region
from sybil.evaluators.skip import evaluate_skip


class AbstractSkipParser:
    """
    An abstract parser for skipping subsequent examples.
    """

    #: Must return :class:`matches<typing.Match>` that contain two groups;
    #: group 1 containing an action of ``'next'``, ``'start'`` or ``'end'`` and group 2
    #: which should contain the source for an optional parenthesis-surrounded Python expression.
    pattern: Pattern

    def __call__(self, document: Document) -> Iterable[Region]:
        for match in re.finditer(self.pattern, document.text):
            yield Region(match.start(), match.end(), match.groups(), evaluate_skip)
